import os
import sys
import logging
from omnik.ha_logger import hybridlogger

from datetime import datetime, timedelta, timezone
import threading

logging.basicConfig(stream=sys.stdout,
                    level=os.environ.get('LOGLEVEL', logging.INFO))

__version__ = '1.0.2'

logger = logging.getLogger(__name__)


class RepeatedJob(object):
    def __init__(self, c, datalogger, hass_api, *args, **kwargs):
        if c.get('default', 'debug', fallback=False):
            logger.setLevel(logging.DEBUG)
        self.logger = logger
        self.hass_api = hass_api
        self.datalogger = datalogger
        self.client = datalogger.client
        self.use_timer = datalogger.client.use_timer
        if self.use_timer:
            self._timer = None
            self.interval = int(c.get('default', 'interval', fallback=360))
            self.half_interval = self.interval/2
            self.retries = 0
            # Trigger the RepeatedJob using a short timer (1s) for the first time so the initialization can finish
            self.calculated_interval = 1
        else:
            self.semaphore = self.client.semaphore
            self.msgevent = self.client.msgevent
        self.args = args
        self.kwargs = kwargs
        # Initialize retry counter
        self.is_running = False
        self.start()

    def function_thread(self):
        return self.function_thread

    # This handler function is fired when the timer has reached the interval
    def _run(self):
        self.is_running = False
        # The function calls DataLogger.process()
        self.last_update_time = self.datalogger.process(*self.args, **self.kwargs)
        # Calculate the new timer interval
        if self.last_update_time:
            # Reset retry counter
            self.retries = 0
            if self.last_update_time <= datetime.now(timezone.utc):
                # If last report time + 2x interval is less than the current time then increase
                self.new_report_expected_at = self.last_update_time + timedelta(seconds=self.interval)
                # Check if we have at least 60 seconds for the next cycle
                if (self.new_report_expected_at + timedelta(seconds=-10) <
                        datetime.now(timezone.utc)):
                    # No recent update of missing update: wait {interval} from now()
                    self.new_report_expected_at = datetime.now(timezone.utc) + \
                        timedelta(seconds=self.interval)
            else:
                # Skipping dark period
                self.new_report_expected_at = self.last_update_time
            self.calculated_interval = (self.new_report_expected_at - datetime.now(timezone.utc)).seconds
        else:
            # An error occured calculate retry interval
            retry_interval = self.half_interval
            i = self.retries
            while i > 0:
                # Double retry interval to avoid to much traffic
                retry_interval *= 2
                i -= 1
            # Increment retry counter maximal interval between retries is half_interval * 2 * 2 * 2 = 4 intervals
            if self.retries < 3:
                self.retries += 1
            # Calculate new report time
            self.new_report_expected_at = datetime.now(timezone.utc) + \
                timedelta(seconds=retry_interval)
            self.calculated_interval = (self.new_report_expected_at - datetime.now(timezone.utc)).seconds
        # Make sure we have at least 15 seconds on the time to prevent a deadlocked timer loop
        if self.calculated_interval < 15:
            self.calculated_interval = 15
        hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                            f"new poll in {self.calculated_interval} seconds at {self.new_report_expected_at.isoformat()}.")
        self.start()

    # This function actual starts the timer
    def start(self):
        if self.use_timer:
            # starting actual timer
            if not self.is_running:
                self._timer = threading.Timer(self.calculated_interval, self._run)
                self._timer.daemon = True
                self._timer.start()
                self.is_running = True
        else:
            # use a listing thread to process
            self.listenthread = threading.Thread(target=self._listen_to_events)
            self.listenthread.start()

    def stop(self):
        if self.use_timer:
            self._timer.cancel()
        else:
            # Stop listening
            self.client.stop()
            # exit event message loop
            self.msgevent.set()
        self.is_running = False

    def _listen_to_events(self):
        self.is_running = True
        while self.is_running:
            self.last_update_time = self.datalogger.process(*self.args, **self.kwargs)



from dsmr_parser.clients import create_dsmr_reader, create_tcp_dsmr_reader
import asyncio
from functools import partial
import threading
from omnik.ha_logger import hybridlogger


class Terminal(object):

    def __init__(self, config, logger, hass_api, terminal_name, dsmr_serial_callback, dsmr_version):
        self.config = config
        self.logger = logger
        self.hass_api = hass_api
        self.terminal_name = terminal_name
        self.dsmr_serial_callback = dsmr_serial_callback
        self.mode = self.config.get(f"dsmr.{self.terminal_name}", 'mode', 'device')
        if self.mode not in ['tcp', 'device']:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "ERROR", f"DSMR terminal {self.terminal_name} mode {self.mode} is not valid. "
                                "Should be tcp or device. Ignoring DSMR configuration!")

        self.device = self.config.get(f"dsmr.{self.terminal_name}", 'device', '/dev/ttyUSB0')
        self.host = self.config.get(f"dsmr.{self.terminal_name}", 'host', 'localhost')
        self.port = self.config.get(f"dsmr.{self.terminal_name}", 'port', '3333')
        self.dsmr_version = dsmr_version
        # start terminal
        self.stop = False
        self.transport = None
        hybridlogger.ha_log(self.logger, self.hass_api,
                            "INFO", f"Initializing DSMR termimal '{terminal_name}'. Mode: {self.mode}.")
        self.thr = threading.Thread(target=self._run_terminal, name=self.terminal_name)
        self.thr.start()

    def terminate(self):
        # cleanup connection after user initiated shutdown
        if self.transport:
            self.transport.close()
        self.stop = True
        self.thr.join()

    def _run_terminal(self):
        # SERIAL_SETTINGS_V2_2, SERIAL_SETTINGS_V4, SERIAL_SETTINGS_V5
        # Telegram specifications:
        # V2_2, V3=V2_2, V4, V5, BELGIUM_FLUVIUS, LUXEMBOURG_SMARTY
        # self.log("DSRM parser was started")
        # The serial approach
        # serial_reader = SerialReader(
        # device='/dev/ttyUSB0',
        # serial_settings=SERIAL_SETTINGS_V5,
        # telegram_specification=telegram_specifications.V5
        # )
        loop = asyncio.new_event_loop()
        try:
            if self.mode == 'tcp':
                create_connection = partial(create_tcp_dsmr_reader,
                                            self.host, self.port, self.dsmr_version,
                                            self.dsmr_serial_callback, loop=loop)
                hybridlogger.ha_log(self.logger, self.hass_api,
                                    "INFO", f"Connecting to '{self.host}:{self.port}' using DSMR v{self.dsmr_version}")
            elif self.mode == 'device':
                create_connection = partial(create_dsmr_reader,
                                            self.device, self.dsmr_version,
                                            self.dsmr_serial_callback, loop=loop)
                hybridlogger.ha_log(self.logger, self.hass_api,
                                    "INFO", f"Connecting to '{self.device}' using DSMR v{self.dsmr_version}")
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "ERROR", f"DSMR terminal {self.terminal_name} could not be initialized. "
                                f"Check your configuration. Error: {e}")
            return
        # create serial or tcp connection
        while not self.stop:
            try:
                conn = create_connection()
                self.transport, protocol = loop.run_until_complete(conn)
                # wait until connection it closed
                loop.run_until_complete(protocol.wait_closed())
                # wait 5 seconds before attempting reconnect
                if not self.stop:
                    loop.run_until_complete(asyncio.sleep(5))
            except Exception as e:
                hybridlogger.ha_log(self.logger, self.hass_api,
                                    "WARNING", f"DSMR terminal {self.terminal_name} was interrupted "
                                    f"and will be restarted in a few moments: {e.args}"
                                    )
                asyncio.sleep(10)
        # Close the event loop
        loop.close()



import json
import time
from omnik.ha_logger import hybridlogger

import urllib.parse
import requests
from omnik.plugin_output import Plugin
from decimal import Decimal
import threading


class pvoutput(Plugin):

    def __init__(self):
        super().__init__()
        self.name = 'pvoutput'
        self.description = 'Write output to PVOutput'
        # Enable support for aggregates
        self.process_aggregates = True
        # Make instance to run exclusively
        self.access = threading.Condition(threading.Lock())

    def _get_temperature(self, msg, data):
        if self.config.getboolean('output.pvoutput', 'use_temperature', fallback=True) and \
                self.config.getboolean('output.pvoutput', 'use_inverter_temperature', fallback=False):
            if 'inverter_temperature' in msg:
                data['v5'] = float(msg['inverter_temperature'])
            return
        if self.config.getboolean('output.pvoutput', 'use_temperature', fallback=False) and \
                not self.config.getboolean('output.pvoutput', 'use_inverter_temperature', fallback=False):
            weather = self.get_weather()
            data['v5'] = weather['main']['temp']

    def _get_voltage(self, msg, data):
        voltage_field = self.config.get('output.pvoutput', 'publish_voltage', fallback=None)
        net_voltage_field = self.config.get('output.pvoutput', 'net_voltage_fallback', fallback=None)
        if voltage_field and voltage_field in msg:
            data['v6'] = float(msg[voltage_field])
        elif net_voltage_field and net_voltage_field in msg:
            data['v6'] = float(msg[net_voltage_field])

    def _check_requirements(self, msg):
        if not self.config.has_option('output.pvoutput', 'api_key'):
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR",
                                f'[{__name__}] No api_key found in configuration')
            return False

        if 'sys_id' not in msg:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR",
                                f'[{__name__}] No sys_id found in dataset, set sys_id under '
                                'your plant settings or global under the output.pvoutput section')
            return False
        return True

    def process(self, **args):
        """
        Send data to pvoutput
        """
        try:
            self.access.acquire()

            msg = args['msg']
            reporttime = time.localtime(msg['last_update'])

            # self.logger.debug(json.dumps(msg, indent=2))
            if not self._check_requirements(msg):
                return

            headers = {
                "X-Pvoutput-Apikey": self.config.get('output.pvoutput', 'api_key'),
                "X-Pvoutput-SystemId": str(msg['sys_id']),
                "Content-type": "application/x-www-form-urlencoded",
                "Accept": "text/plain"
            }

            # see: https://pvoutput.org/help.html
            # see: https://pvoutput.org/help.html#api-addstatus
            # report cumulative values to enable power usage
            data = {
                'd': time.strftime('%Y%m%d', reporttime),
                't': time.strftime('%H:%M', reporttime),
                'c1': 0
                }
            # v1 = energy_generated (Wh) * 1000 ; this value is on a dayly basis
            # v2 = bruto power_generated (W)
            # v3 = energy_used (Wh) * 1000 ; this value is cumulative
            # v4 = bruto power_consumption (W)
            # c1 = 1 ; v3 is cumulative so the c1 flag is set
            # see note about Cumulative Enery at https://pvoutput.org/help.html#api-addstatus

            if 'energy_used' in msg and 'power_consumption' in msg:
                data.update({
                    'v3': f"{msg['energy_used'] * Decimal('1000')}",
                    'v4': f"{msg['power_consumption']}",
                    'c1': 1
                    })
                # v3 = energy_used (Wh) * 1000 ; this value is cumulative
                # v4 = bruto power_consumption (W)
                # c1 = 1 ; v3 is cumulative so the c1 flag is set
                # see note about Cumulative Enery at https://pvoutput.org/help.html#api-addstatus
            if 'current_power' in msg:
                data.update({
                    'v1': f"{msg['today_energy'] * Decimal('1000')}",
                    'v2': f"{msg['current_power']}",
                    })
                # v1 = energy_generated (Wh) * 1000 ; this value is on a dayly basis
                # v2 = bruto power_generated (W)
                # c1 = 0 ; v1 is not cumulative so the c1 flag not is set
                # see note about Cumulative Enery at https://pvoutput.org/help.html#api-addstatus

                # Publish (inverter) temperature if solar data is published and a temperature is available.
                # alternatively use the temperature from openweather
                self._get_temperature(msg, data)

            # Publish voltage (if available)
            self._get_voltage(msg, data)

            encoded = urllib.parse.urlencode(data)

            self.logger.debug(json.dumps(data, indent=2))

            r = requests.post(
                "http://pvoutput.org/service/r2/addstatus.jsp", data=encoded, headers=headers)

            r.raise_for_status()

        except requests.exceptions.RequestException as err:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"Unhandled request error: {err}")
        except requests.exceptions.HTTPError as errh:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"Http error: {errh}")
        except requests.exceptions.ConnectionError as errc:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"Connection error: {errc}")
        except requests.exceptions.Timeout as errt:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"Timeout error: {errt}")
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", f"Unexpected error: {e.args}")
        finally:
            self.access.release()



import requests
from cachetools import TTLCache
from omnik.ha_logger import hybridlogger
from decimal import Decimal

# TODO: Create Abstract Base Class


class BasePlugin(type):
    def __init__(cls, name, bases, attrs):
        super(BasePlugin, cls).__init__(name)
        if not hasattr(cls, 'plugins'):
            cls.plugins = []
        else:
            cls.register(cls)  # Called when a plugin class is imported

    def register(cls, plugin):
        cls.plugins.append(plugin())


class Plugin(object, metaclass=BasePlugin):

    config = None
    logger = None
    hass_api = None
    process_aggregates = False

    cache = TTLCache(maxsize=1, ttl=300)

    def jsonval(self, value):
        if isinstance(value, Decimal):
            if float(value) == int(value):
                return int(value)
            else:
                return float(value)
        elif isinstance(value, float):
            return value
        elif isinstance(value, int):
            return value
        elif isinstance(value, bool):
            return value
        elif isinstance(value, str):
            return value
        else:
            return str(value)

    def get_weather(self):
        try:
            if 'weather' not in self.cache:
                self.logger.debug('[cache miss] Fetching weather data')
                url = "https://{endpoint}/data/2.5/weather?lon={lon}&lat={lat}&units={units}&APPID={api_key}".format(
                    endpoint=self.config.get('openweathermap', 'endpoint', fallback='api.openweathermap.org'),
                    lat=self.config.get('openweathermap', 'lat'),
                    lon=self.config.get('openweathermap', 'lon'),
                    units=self.config.get(
                        'openweathermap', 'units', fallback='metric'),
                    api_key=self.config.get('openweathermap', 'api_key'),
                )

                res = requests.get(url)

                res.raise_for_status()

                self.cache['weather'] = res.json()

            return self.cache['weather']

        except requests.exceptions.HTTPError as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", 'Unable to get weather data. [{0}]: {1}'.format(
                                type(e).__name__, str(e)))
            raise e

    def process(self, **args):
        pass



import requests
import urllib.parse
from omnik.ha_logger import hybridlogger
import hashlib
import xml.etree.ElementTree as ET
import datetime
from omnik.plugin_client import Client


class SolarmanPVClient(Client):

    def __init__(self):
        super().__init__()
        hybridlogger.ha_log(self.logger, self.hass_api, "INFO", "Client enabled: solarmanpvclient")

        # API Key
        self.api_key = self.config.get('client.solarmanpv', 'api_key', fallback='apitest')
        self.user_id = -1
        self.token = None

        self.base_url = self.config.get('client.solarmanpv', 'base_url',
                                        fallback='http://www.solarmanpv.com:18000/SolarmanApi/serverapi')

        self.username = self.config.get('client.solarmanpv', 'username')
        self.password = self.config.get('client.solarmanpv', 'password')

    def initialize(self):
        pwhash = hashlib.md5(self.password.encode('utf-8')).hexdigest()
        url = f'{self.base_url}/Login?username={self.username}&password={pwhash}&key={self.api_key}'

        xmldata = self._api_request(url, 'GET', None)
        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", f"account validation: {xmldata}")

        # Parsing the XML
        for element in ET.fromstring(xmldata):
            if element.tag == "userID":
                self.user_id = int(element.text)
            elif element.tag == "token":
                self.token = element.text

    def _api_request(self, url, method, body, encode=False):
        '''
        This API mithod uses XML. When logging in a token is obtained.
        For each request is to be checked if the token is still valid.
        '''
        headers = {
            'Content-type': 'application/x-www-form-urlencoded'
        }

        data = urllib.parse.urlencode(
            body) if encode and body is not None else body

        r = requests.request(
            method=method,
            url=url,
            data=data,
            headers=headers
        )

        r.raise_for_status()

        return r.content

    def getPlants(self):

        data = []
        for plant in self.config.getlist('client.solarmanpv', 'plant_id_list'):
            data.append({'plant_id': plant})

        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", f"plant list from config {data}")

        return data

    def xmlprop(self, xml, params, fallback=None):
        paramlist = params
        param = paramlist.pop(0)
        for el in xml:
            if param == el.tag:
                if paramlist:
                    return self.xmlprop(el, paramlist, fallback)
                else:
                    return el.text
        return fallback

    def getPlantData(self, plant_id):

        data = {}
        url = f'{self.base_url}/Data?username={self.username}&stationid={plant_id}&key={self.api_key}&token={self.token}'

        xmldata = self._api_request(url, 'GET', None)
        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", f"plant data ({plant_id}) {xmldata}")
        # Parsing the XML
        xml = ET.fromstring(xmldata)

        # Check if an error was returned
        if xml.tag != 'data':
            # log warning
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", "No valid data received. Trying to renew token.")
            # Get new token
            self.initialize()
            # Retry the call
            url = f'{self.base_url}/Data?username={self.username}&stationid={plant_id}&key={self.api_key}&token={self.token}'
            xmldata = self._api_request(url, 'GET', None)
            hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", f"plant data retry ({plant_id}) {xmldata}")
            # Parsing the XML
            xml = ET.fromstring(xmldata)

        if xml.tag == 'data':
            data['name'] = self.xmlprop(xml, ['name'])
            data['income'] = float(self.xmlprop(xml, ['income', 'TotalIncome'], 0.0))
            data['data_logger'] = self.xmlprop(xml, ['detail', 'WiFi', 'id'], '')
            data['inverter'] = self.xmlprop(xml, ['detail', 'WiFi', 'inverter', 'SN'], '')
            data['status'] = int(self.xmlprop(xml, ['detail', 'WiFi', 'inverter', 'status'], 0))
            data['current_power'] = int(float(self.xmlprop(xml, ['detail', 'WiFi', 'inverter', 'power'], 0.0)) * 1000)
            data['today_energy'] = float(self.xmlprop(xml, ['detail', 'WiFi', 'inverter', 'etoday'], 0.0))
            data['total_energy'] = float(self.xmlprop(xml, ['detail', 'WiFi', 'inverter', 'etotal'], 0.0))
            data['last_update'] = float(self.xmlprop(xml, ['detail', 'WiFi', 'inverter', 'lastupdated'],
                                        datetime.datetime.now().timestamp()))
            data['last_update_time'] = datetime.datetime.utcfromtimestamp(data['last_update']).strftime('%Y-%m-%dT%H:%M:%SZ')
            return data
        else:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", f"Cannot fetch data: {xmldata}")
            return None



import datetime
import pytz
from astral import Astral

global default_timezone
global default_city_name

default_timezone = 'CET'
default_city_name = 'Amsterdam'


class daylight(object):

    def __init__(self, city_name=default_city_name):
        self._a = Astral()
        self._a.solar_depression = 'civil'
        self._city = self._a[city_name]
        self._timezone = pytz.timezone(self._city.timezone)

    @property
    def dawn(self, date=None):
        if not date:
            date = datetime.datetime.now(self.timezone)
        self.sun = self._city.sun(date=date)
        return self.sun['dawn']

    @property
    def next_dawn(self, date=None):
        if not date:
            date = datetime.datetime.now(self.timezone)
        self.sun = self._city.sun(date=date)
        if (self.sun['dawn'] < date):
            self.sun = self._city.sun(date=date+datetime.timedelta(days=1))
        return self.sun['dawn']

    @property
    def sunrise(self, date=None):
        if not date:
            date = datetime.datetime.now(self.timezone)
        self.sun = self._city.sun(date=date)
        return self.sun['sunrise']

    @property
    def noon(self, date=None):
        self.sun = self._city.sun(date=date)
        return self.sun['noon']

    @property
    def sunset(self, date=None):
        if not date:
            date = datetime.datetime.now(self.timezone)
        self.sun = self._city.sun(date=date)
        return self.sun['sunset']

    @property
    def dusk(self, date=None):
        if not date:
            date = datetime.datetime.now(self.timezone)
        self.sun = self._city.sun(date=date)
        return self.sun['dusk']

    @property
    def timezone(self):
        return self._timezone

    def sun_rising(self, time: datetime.datetime = None):
        if not time:
            time = datetime.datetime.now(self.timezone)
        self.sun = self._city.sun(date=time)
        return (time > self.sun['dawn']) and (time < self.sun['sunrise'])

    def sun_up(self, time: datetime.datetime = None):
        if not time:
            time = datetime.datetime.now(self.timezone)
        self.sun = self._city.sun(date=time)
        return (time > self.sun['sunrise']) and (time < self.sun['sunset'])

    def sun_shine(self, time: datetime.datetime = None):
        if not time:
            time = datetime.datetime.now(self.timezone)
        self.sun = self._city.sun(date=time)
        return (time > self.sun['dawn']) and (time < self.sun['dusk'])

    def sun_setting(self, time: datetime.datetime = None):
        if not time:
            time = datetime.datetime.now(self.timezone)
        self.sun = self._city.sun(date=time)
        return (time > self.sun['sunset']) and (time < self.sun['dusk'])

    def sun_down(self, time: datetime.datetime = None):
        if not time:
            time = datetime.datetime.now(self.timezone)
        self.sun = self._city.sun(date=time)
        return (time < self.sun['dawn']) or (time > self.sun['dusk'])


def main():
    dl = daylight('Amsterdam')
    nu = datetime.datetime.now(tz=dl.timezone)
    print('Now:     %s' % str(nu))
    print('Timezone:%s' % str(dl.timezone))
    print('Dawn:    %s' % str(dl.dawn))
    print('nxtdawn: %s' % str(dl.next_dawn))
    print('Sunrise: %s' % str(dl.sunrise))
    print('Noon:    %s' % str(dl.noon))
    print('Sunset:  %s' % str(dl.sunset))
    print('Dusk:    %s' % str(dl.dusk))
    print('dDawn:    %s' % str(dl.dawn-nu))
    print('dSunrise: %s' % str(dl.sunrise-nu))
    print('dNoon:    %s' % str(dl.noon-nu))
    print('dSunset:  %s' % str(dl.sunset-nu))
    print('dDusk:    %s' % str(dl.dusk-nu))
    print('Sun rising :', dl.sun_rising())
    print('Sun up     :', dl.sun_up())
    print('Sun shine  :', dl.sun_shine())
    print('Sun setting:', dl.sun_setting())
    print('Sun down   :', dl.sun_down())


if __name__ == "__main__":
    main()



import requests
import urllib.parse
from omnik.ha_logger import hybridlogger
from omnik.plugin_client import Client
from datetime import datetime


class OmnikPortalClient(Client):

    def __init__(self):
        super().__init__()
        hybridlogger.ha_log(self.logger, self.hass_api, "INFO", "Client enabled: omnikportalclient")

        # API Key's
        self.app_id = self.config.get('client.omnikportal', 'app_id', fallback='10038')
        self.app_key = self.config.get('client.omnikportal', 'app_key', fallback='Ox7yu3Eivicheinguth9ef9kohngo9oo')
        self.user_id = -1

        self.base_url = self.config.get('client.omnikportal', 'base_url', fallback='https://api.omnikportal.com/v1')

        self.username = self.config.get('client.omnikportal', 'username')
        self.password = self.config.get('client.omnikportal', 'password')

    def initialize(self):
        url = f'{self.base_url}/user/account_validate'

        body = {
            'user_email': self.username,
            'user_password': self.password,
            'user_type': 1
        }

        data = self._api_request(url, 'POST', body)
        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", f"account validation: {data}")

        # this is what `initialize` does ... setting the `user_id`
        self.user_id = data['data']['c_user_id']

    def _api_request(self, url, method, body, encode=False):
        headers = {
            'uid': str(self.user_id),
            'app_id': str(self.app_id),
            'app_key': self.app_key,
            'Content-type': 'application/x-www-form-urlencoded'
        }

        data = urllib.parse.urlencode(
            body) if encode and body is not None else body

        r = requests.request(
            method=method,
            url=url,
            data=data,
            headers=headers
        )

        r.raise_for_status()

        return r.json()

    def getPlants(self):
        url = f'{self.base_url}/plant/list'

        data = self._api_request(url, 'GET', None)
        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", f"plant list {data}")

        return data['data'].get('plants', [])

    def getPlantData(self, plant_id):

        url = f'{self.base_url}/plant/data?plant_id={plant_id}'

        rawdata = self._api_request(url, 'GET', None)
        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", f"plant data ({plant_id}) {rawdata}")
        data = rawdata['data']
        # Adjust power to Watt (not kW) and covert strings numbers to float or int
        data['total_energy'] = float(data['total_energy'])
        data['current_power'] = int(data['current_power'] * 1000)
        data['peak_power_actual'] = int(data['peak_power_actual'] * 1000)
        data['income'] = float(data['income'])
        data['last_update'] = datetime.timestamp(datetime.strptime(f"{data['last_update_time']} UTC+0000",
                                                 '%Y-%m-%dT%H:%M:%SZ %Z%z'))

        return data



#! /usr/bin/env python3

import os
import argparse
import configparser
import pathlib
import time
import logging
import appdaemon.plugins.hass.hassapi as hass
from omnik.ha_logger import hybridlogger
from omnik import RepeatedJob
from omnik.datalogger import DataLogger

logger = logging.getLogger(__name__)

# customize config parser with dict based lookup for AppDaemon and command line options
# has_option and get function have been adapted
ha_args = {}


class ha_ConfigParser(configparser.ConfigParser):

    def __init__(self, ha_args: dict = None, *args, **kwargs):
        if ha_args is None:
            self.ha_args = {}
        else:
            self.ha_args = ha_args
        super().__init__(*args, **kwargs)

    def has_option(self, section: str, option: str):
        if str(section).lower() == 'default':
            if option in self.ha_args:
                return True
        else:
            if section in self.ha_args:
                if option in self.ha_args[section]:
                    return True
        return super().has_option(section, option)

    def get(self, section, option, fallback=None, **kwargs):
        if str(section).lower() == 'default':
            if option in self.ha_args:
                return self.ha_args.get(option, fallback)
        else:
            if str(section) in self.ha_args:
                if option in self.ha_args[section]:
                    return self.ha_args[section].get(option, fallback)
        try:
            if fallback:
                retval = super().get(section, option, fallback=fallback, **kwargs)
            else:
                retval = super().get(section, option, fallback='', **kwargs)
        except Exception:
            retval = fallback
        return retval

    def getboolean(self, section, option, fallback=None, **kwargs):
        truelist = ['true', '1', 't', 'y', 'yes', 'j', 'ja']
        valstr = str(self.get(section, option, fallback, **kwargs)).lower()
        return (valstr in truelist)

    def getlist(self, section, option, fallback=[], **kwargs):
        if str(section).lower() == 'default':
            if option in self.ha_args:
                return self.ha_args.get(option, fallback)
        else:
            if str(section) in self.ha_args:
                if option in self.ha_args[section]:
                    return self.ha_args[section].get(option, fallback)
        try:
            retval = super().getlist(section, option, fallback=fallback, **kwargs)
        except Exception:
            retval = fallback
            pass
        return retval


# Initialization class for AppDaemon (homeassistant)
class HA_OmnikDataLogger(hass.Hass):

    def initialize(self, *args, **kwargs):
        hybridlogger.ha_log(logger, self, "INFO", "Starting Omnik datalogger...")
        hybridlogger.ha_log(logger, self, "DEBUG", f"Arguments from AppDaemon config (self.args): {self.args}")
        if 'config' in self.args:
            c = ha_ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]}, ha_args=self.args)
            self.configfile = self.args['config']
            try:
                c.read(self.configfile, encoding='utf-8')
                c.configfile = self.configfile
            except Exception as e:
                hybridlogger.ha_log(logger, self, "ERROR",
                                    f"Error parsing 'config.ini' from {self.configfile}. No valid configuration file. {e}.")
                c.configfile = None
        else:
            c = ha_ConfigParser(ha_args=self.args)
            c.configfile = None
        set_data_config_path(c)

        self.interval = int(c.get('default', 'interval', 360))
        self.datalogger = DataLogger(c, hass_api=self)

        # running repeatedly
        if self.datalogger.client.use_timer:
            hybridlogger.ha_log(logger, self, "INFO", f"Daemon interval {self.interval} seconds.")
        else:
            hybridlogger.ha_log(logger, self, "INFO", "Daemon starts listening.")
        self.rt = RepeatedJob(c, self.datalogger, hass_api=self)

    def terminate(self):
        self.datalogger.terminate()
        hybridlogger.ha_log(logger, self, "INFO", "DSMR Daemon was stopped.")
        self.rt.stop()
        hybridlogger.ha_log(logger, self, "INFO", "Daemon was stopped.")


# Initialization from the commandline
def main(c: ha_ConfigParser, hass_api=None):
    # Enabled debugging if the flag is set
    if c.get('default', 'debug', False):
        logger.setLevel(logging.DEBUG)
    set_data_config_path(c)
    datalogger = DataLogger(c, hass_api=hass_api)
    if c.has_option('default', 'interval'):
        # running repeatedly
        if datalogger.client.use_timer:
            hybridlogger.ha_log(logger, hass_api, "INFO", f"Daemon interval {int(c.get('default', 'interval', 360))} seconds.")
        else:
            hybridlogger.ha_log(logger, hass_api, "INFO", "Daemon starts listening.")
        rt = RepeatedJob(c, datalogger, hass_api)
        if hass_api:
            hass_api.rt = rt
        else:
            # loop is for commandline execution only
            try:
                while True:
                    time.sleep(0.5)
            except KeyboardInterrupt:
                hybridlogger.ha_log(logger, hass_api, "INFO", '> stopping all threads')
                rt.stop()
                datalogger.terminate()

    else:
        # running only once
        datalogger.process()
        datalogger.terminate()


def set_data_config_path(config):
    config.data_config_file_path = f"{pathlib.Path(__file__).parent.absolute()}/data_fields.json"
    config.data_config_file_shared = f"{pathlib.Path(__file__).parent.parent.absolute()}" \
                                     "/share/omnikdatalogger/data_fields.json"


if __name__ == '__main__':
    home = os.path.expanduser('~')
    parser = argparse.ArgumentParser()

    parser.add_argument('--config', default=os.path.join(home, '.omnik/config.ini'),
                        help='Path to configuration file', metavar="FILE")
    parser.add_argument('--interval', type=int, help='Execute every n seconds')
    parser.add_argument('-d', '--debug', action='store_true', help='Debug mode')

    args = parser.parse_args()

    if not os.path.isfile(args.config):
        logging.critical("Unable to find config at '{}'".format(args.config))
        parser.print_help()
        os.sys.exit(1)
    if args.config:
        ha_args['config'] = args.config
        if args.debug:
            ha_args['debug'] = args.debug
        if args.interval:
            ha_args['interval'] = args.interval
        c = ha_ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]}, ha_args=ha_args)
        c.read([args.config], encoding='utf-8')
        c.configfile = args.config

    main(c, hass_api=None)



from dsmr_parser import obis_references
from .terminal import Terminal
from decimal import Decimal
import threading
from datetime import datetime
from omnik.ha_logger import hybridlogger


class DSRM(object):

    config = None
    logger = None
    hass_api = None
    datalogger = None

    def initialize(self, terminals, dsmr_callback):
        # args.host, args.port, args.version, args.device
        self.dsmr_callback = dsmr_callback
        # self.stop = False

        self._init_terminals(terminals)

    def _init_terminals(self, terminals):
        self.terminals = {}
        self.sync = {}
        self.cache = {}
        self.ts_last_telegram = {}
        self.last_gas_update = {}
        self.tconfig = {}
        self.tconfig['tarif'] = {}
        if self.config.has_option('dsmr', 'tarif'):
            tarif_list = self.config.getlist('dsmr', 'tarif')
        else:
            tarif_list = ['0001', '0002']
            self.tconfig['tarif']['0001'] = 'low'
            self.tconfig['tarif']['0002'] = 'normal'
        for tarif in tarif_list:
            if tarif in self.tconfig['tarif']:
                default = self.tconfig['tarif'][tarif]
            else:
                default = tarif
            self.tconfig['tarif'][tarif] = self.config.get('dsmr', f"tarif.{tarif}", default)
        for terminal in terminals:
            # Pre read configuration for each terminal
            self.tconfig[terminal] = {}

            self.tconfig[terminal]['plant_id'] = self.config.get(f"dsmr.{terminal}", 'plant_id', '')
            self.tconfig[terminal]['gas_meter'] = self.config.getboolean(f"dsmr.{terminal}", 'gas_meter', True)
            self.tconfig[terminal]['dsmr_version'] = self.config.get(f"dsmr.{terminal}", 'dsmr_version ', '5')
            self.tconfig[terminal]['total_energy_offset'] = \
                Decimal(self.config.get(f"dsmr.{terminal}", 'total_energy_offset', '0'))

            # Init terminal sync parameters
            self.sync[terminal] = 0
            self.cache[terminal] = Decimal(1000000)
            self.ts_last_telegram[terminal] = 0
            self.last_gas_update[terminal] = [0, Decimal('0.0'), Decimal('0.000')]

            # Warnings and errors
            if not self.tconfig[terminal]['plant_id']:
                hybridlogger.ha_log(self.logger, self.hass_api,
                                    "WARNING", f"DSMR 'plant_id' for terminal '{terminal}' is not specified. "
                                    "Received data will be NOT be associated with your solar data "
                                    "and will be processed as stand-a-lone data!")
            # Initialize terminal
            self.terminals[terminal] = Terminal(self.config, self.logger, self.hass_api, terminal,
                                                self.dsmr_serial_callback, self.tconfig[terminal]['dsmr_version'])

    def terminate(self):
        # cleanup connection after user initiated shutdown
        for terminal in self.terminals:
            self.terminals[terminal].terminate()

    def _proces_power_current_fase(self, fase, msg_dsmr, telegram):

        # Build reference arrays
        voltage_fase_obis = [obis_references.INSTANTANEOUS_VOLTAGE_L1,
                             obis_references.INSTANTANEOUS_VOLTAGE_L2,
                             obis_references.INSTANTANEOUS_VOLTAGE_L3]
        voltage_fase_sag_obis = [obis_references.VOLTAGE_SAG_L1_COUNT,
                                 obis_references.VOLTAGE_SAG_L2_COUNT,
                                 obis_references.VOLTAGE_SAG_L3_COUNT]
        voltage_fase_swell_obis = [obis_references.VOLTAGE_SAG_L1_COUNT,
                                   obis_references.VOLTAGE_SAG_L2_COUNT,
                                   obis_references.VOLTAGE_SAG_L3_COUNT]
        current_fase_obis = [obis_references.INSTANTANEOUS_CURRENT_L1,
                             obis_references.INSTANTANEOUS_CURRENT_L2,
                             obis_references.INSTANTANEOUS_CURRENT_L3]
        pwr_pos_fase_obis = [obis_references.INSTANTANEOUS_ACTIVE_POWER_L1_POSITIVE,
                             obis_references.INSTANTANEOUS_ACTIVE_POWER_L2_POSITIVE,
                             obis_references.INSTANTANEOUS_ACTIVE_POWER_L3_POSITIVE]
        pwr_neg_fase_obis = [obis_references.INSTANTANEOUS_ACTIVE_POWER_L1_NEGATIVE,
                             obis_references.INSTANTANEOUS_ACTIVE_POWER_L2_NEGATIVE,
                             obis_references.INSTANTANEOUS_ACTIVE_POWER_L3_NEGATIVE]

        sfase = f"{fase}"

        if pwr_pos_fase_obis[fase - 1] in telegram:
            msg_dsmr[f'INSTANTANEOUS_CURRENT_L{sfase}'] = telegram[current_fase_obis[fase - 1]].value
            msg_dsmr[f'VOLTAGE_SAG_L{sfase}_COUNT'] = telegram[voltage_fase_sag_obis[fase - 1]].value
            msg_dsmr[f'VOLTAGE_SWELL_L{sfase}_COUNT'] = telegram[voltage_fase_swell_obis[fase - 1]].value
            msg_dsmr[f'INSTANTANEOUS_ACTIVE_POWER_L{fase}_POSITIVE'] = telegram[pwr_pos_fase_obis[fase - 1]].value
            msg_dsmr[f'INSTANTANEOUS_ACTIVE_POWER_L{fase}_NEGATIVE'] = telegram[pwr_neg_fase_obis[fase - 1]].value
            # current_net_power_l{fase} is calculated in Watt and is negative when power is delivered
            msg_dsmr[f'current_net_power_l{sfase}'] = ((msg_dsmr[f'INSTANTANEOUS_ACTIVE_POWER_L{sfase}_POSITIVE'] -
                                                        msg_dsmr[f'INSTANTANEOUS_ACTIVE_POWER_L{sfase}_NEGATIVE']) *
                                                       Decimal('1000')).quantize(Decimal('1.'))
        if voltage_fase_obis[fase - 1] in telegram:
            # We have voltage, so lets calculate the current using power and voltage
            msg_dsmr[f'INSTANTANEOUS_VOLTAGE_L{sfase}'] = telegram[voltage_fase_obis[fase - 1]].value
            # current_net_current_l{fase} is calculated in Ampère and is negative when power is delivered
            msg_dsmr[f'net_current_l{sfase}'] = (msg_dsmr[f'current_net_power_l{sfase}'] /
                                                 msg_dsmr[f'INSTANTANEOUS_VOLTAGE_L{sfase}']).quantize(Decimal('.01'))

    def _process_power_details(self, msg_dsmr, telegram):
        terminal = threading.currentThread().getName()
        try:
            # Get global power (kWh)
            msg_dsmr['CURRENT_ELECTRICITY_USAGE'] = telegram[obis_references.CURRENT_ELECTRICITY_USAGE].value
            msg_dsmr['CURRENT_ELECTRICITY_DELIVERY'] = telegram[obis_references.CURRENT_ELECTRICITY_DELIVERY].value
            # Get energy usage
            msg_dsmr['ELECTRICITY_USED_TARIFF_1'] = telegram[obis_references.ELECTRICITY_USED_TARIFF_1].value
            msg_dsmr['ELECTRICITY_USED_TARIFF_2'] = telegram[obis_references.ELECTRICITY_USED_TARIFF_2].value
            msg_dsmr['energy_used_net'] = msg_dsmr['ELECTRICITY_USED_TARIFF_1'] + \
                msg_dsmr['ELECTRICITY_USED_TARIFF_2']  # ELECTRICITY_USED (NV)
            msg_dsmr['ELECTRICITY_DELIVERED_TARIFF_1'] = telegram[obis_references.ELECTRICITY_DELIVERED_TARIFF_1].value
            msg_dsmr['ELECTRICITY_DELIVERED_TARIFF_2'] = telegram[obis_references.ELECTRICITY_DELIVERED_TARIFF_2].value
            msg_dsmr['energy_delivered_net'] = msg_dsmr['ELECTRICITY_DELIVERED_TARIFF_1'] + \
                msg_dsmr['ELECTRICITY_DELIVERED_TARIFF_2']
            tarif = telegram[obis_references.ELECTRICITY_ACTIVE_TARIFF].value
            if tarif in self.tconfig['tarif']:
                msg_dsmr['ELECTRICITY_ACTIVE_TARIFF'] = self.tconfig['tarif'][tarif]
            else:
                msg_dsmr['ELECTRICITY_ACTIVE_TARIFF'] = tarif

            if self.tconfig[terminal]['dsmr_version'] in ['5B', '5', '4']:
                msg_dsmr['LONG_POWER_FAILURE_COUNT'] = telegram[obis_references.LONG_POWER_FAILURE_COUNT].value
                msg_dsmr['SHORT_POWER_FAILURE_COUNT'] = telegram[obis_references.SHORT_POWER_FAILURE_COUNT].value

                for fase in [1, 2, 3]:
                    self._proces_power_current_fase(fase, msg_dsmr, telegram)
                net_voltage_max = Decimal('0')
                if 'INSTANTANEOUS_VOLTAGE_L1' in msg_dsmr:
                    net_voltage_max = max([msg_dsmr['INSTANTANEOUS_VOLTAGE_L1'], net_voltage_max])
                if 'INSTANTANEOUS_VOLTAGE_L2' in msg_dsmr:
                    net_voltage_max = max([msg_dsmr['INSTANTANEOUS_VOLTAGE_L2'], net_voltage_max])
                if 'INSTANTANEOUS_VOLTAGE_L3' in msg_dsmr:
                    net_voltage_max = max([msg_dsmr['INSTANTANEOUS_VOLTAGE_L3'], net_voltage_max])
                if net_voltage_max:
                    msg_dsmr['net_voltage_max'] = net_voltage_max

            # msg_dsmr['current_net_power'] = Decimal('0')
            # for fase in [1,2,3]:
            #    if f'current_net_power_l{fase}' in msg_dsmr:
            #        msg_dsmr['current_net_power'] += msg_dsmr[f'current_net_power_l{fase}']
            # if not msg_dsmr['current_net_power']:
            # Use global power indicators
            msg_dsmr['current_net_power'] = ((msg_dsmr['CURRENT_ELECTRICITY_USAGE'] -
                                              msg_dsmr['CURRENT_ELECTRICITY_DELIVERY']) *
                                             Decimal('1000')).quantize(Decimal('1.'))

        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "WARNING", f"DSMR current calculation for terminal '{terminal}' failed. "
                                f"Check your config: Error: {e.args}")

    def _process_gas(self, msg_dsmr, telegram):
        terminal = threading.currentThread().getName()
        if not self.tconfig[terminal]['gas_meter']:
            # Skip gas meter
            return
        try:
            if self.tconfig[terminal]['dsmr_version'] in ['4', '5']:
                G = telegram[obis_references.HOURLY_GAS_METER_READING]
                msg_dsmr['gas_consumption_total'] = G.values[1]['value']
                msg_dsmr['timestamp_gas'] = datetime.timestamp(G.values[0]['value'])
            elif self.tconfig[terminal]['dsmr_version'] == '5B':
                G = telegram[obis_references.BELGIUM_HOURLY_GAS_METER_READING]
                msg_dsmr['gas_consumption_total'] = G.values[1]['value']
                msg_dsmr['timestamp_gas'] = datetime.timestamp(G.values[0]['value'])
            elif self.tconfig[terminal]['dsmr_version'] == '2.2':
                G = telegram[obis_references.GAS_METER_READING]
                msg_dsmr['gas_consumption_total'] = G.values[6]['value']
                msg_dsmr['timestamp_gas'] = datetime.timestamp(G.values[0]['value'])
            msg_dsmr['EQUIPMENT_IDENTIFIER_GAS'] = telegram[obis_references.EQUIPMENT_IDENTIFIER_GAS].value
            # Set last value to the cache
            if not self.last_gas_update[terminal][0]:
                self.last_gas_update[terminal][0] = msg_dsmr['timestamp_gas']
                self.last_gas_update[terminal][1] = msg_dsmr['gas_consumption_total']
            # Calculate gas consumption / hour self.last_gas_update[terminal] = [0, Decimal('0')]
            if msg_dsmr['timestamp_gas'] > self.last_gas_update[terminal][0]:
                msg_dsmr['gas_consumption_hour'] = ((msg_dsmr['gas_consumption_total'] -
                                                    self.last_gas_update[terminal][1]) * Decimal('3600') /
                                                    Decimal(msg_dsmr['timestamp_gas'] - self.last_gas_update[terminal][0])
                                                    ).quantize(Decimal('0.001'))
                self.last_gas_update[terminal][0] = msg_dsmr['timestamp_gas']
                self.last_gas_update[terminal][1] = msg_dsmr['gas_consumption_total']
                self.last_gas_update[terminal][2] = msg_dsmr['gas_consumption_hour']
            else:
                msg_dsmr['gas_consumption_hour'] = self.last_gas_update[terminal][2]
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "WARNING", f"DSMR gas_meter reading for terminal '{terminal}' failed. "
                                f"Check your config: Error: {e.args}")

    def dsmr_serial_callback(self, telegram):
        terminal = threading.currentThread().getName()
        """Callback that prints telegram values."""
        if obis_references.P1_MESSAGE_TIMESTAMP in telegram:
            # Time stamp from Telegram
            TS = datetime.timestamp(telegram[obis_references.P1_MESSAGE_TIMESTAMP].value)
        else:
            # Create time stamp
            TS = int(datetime.timestamp(datetime.now()))
        # Use power value to sync DSMR 5 telegram flod. DSMR 4 and 2 log possible every 10 sec
        P1 = telegram[obis_references.CURRENT_ELECTRICITY_USAGE]
        P2 = telegram[obis_references.CURRENT_ELECTRICITY_DELIVERY]
        P = ((P2.value - P1.value) * Decimal('1000')).quantize(Decimal('1.'))

        # Sync telegrams if there is a power change of after 10s
        if (self.cache[terminal] != P) or (TS - self.ts_last_telegram[terminal] > 9.5) or (self.sync[terminal] == 10):
            self.cache[terminal] = P
            self.sync[terminal] = 0
            self.ts_last_telegram[terminal] = TS
        # Process telegram if synced
        if self.sync[terminal] == 0:
            # create uniform dsmr message structure for futher processinh
            msg_dsmr = {}
            # Add generic values
            msg_dsmr['timestamp'] = TS
            msg_dsmr['EQUIPMENT_IDENTIFIER'] = telegram[obis_references.EQUIPMENT_IDENTIFIER].value
            msg_dsmr['terminal'] = terminal
            msg_dsmr['plant_id'] = self.tconfig[terminal]['plant_id']
            msg_dsmr['total_energy_offset'] = self.tconfig[terminal]['total_energy_offset']
            # msg_dsmr['advertize_power_fields'] =  self.tconfig[terminal]['advertize_power_fields']
            # msg_dsmr['advertize_dsrm_solar_fields'] =  self.tconfig[terminal]['advertize_dsrm_solar_fields']

            # Read and process more accurate power calculations DSMR 5
            self._process_power_details(msg_dsmr, telegram)

            # Process Gas
            self._process_gas(msg_dsmr, telegram)

            # Send back data to data logger
            self.dsmr_callback(terminal, msg_dsmr)

        # Increase terminal sync counter
        self.sync[terminal] += 1
